@@ -3,7 +3,7 @@
 # e.click()
 # e.send(KEYS.COMMAND,KEYS.SHIFT,KEYS.ARROW_LEFT)
 # e.send(KEYS.COMMAND,'a')
-# e.submit(), senddelay(), sendsubmit()
+# e.submit(), e.sendsubmit()
 # f.find_all(By.CLASS_NAME,'d2l-datetime-selector-date-input.d2l-edit')  
 # https://stackoverflow.com/questions/17000703/is-string-matches-supported-in-selenium-webdriver-2
 # https://stackoverflow.com/questions/22436789/xpath-ends-with-does-not-work
@@ -87,7 +87,7 @@ class Firefox:
         service_log_path: Optional[str] = os.devnull,
         log_path: Optional[str] = os.devnull,
         cookies_folder_path: Optional[str] = None, # not working? Use session instead
-        extensions_folder_path: Optional[str] = None, # ~/Library/Application Support/Firefox/Profiles/h7d5u192.default-esr/extensions
+        extensions_folder_path: Optional[str] = None, # about:support  ~/Library/Application Support/Firefox/Profiles/h7d5u192.default-esr/extensions
         host: Optional[str] = None,
         port: Optional[int] = None,
         cookies_id: Optional[str] = None,
@@ -98,12 +98,13 @@ class Firefox:
         private: bool = False,
         screen_size: Optional[Tuple[int, int]] = None, # (width, height)
         full_screen: bool = False,
+        max_window: bool = True,  # overrides full_screen
         language: str = 'en-us',
         manual_set_timezone: bool = False,
         user_agent: Optional[str] = None,  # random
         load_proxy_checker_website: bool = False,
         disable_images: bool = False,
-        autoclose: bool = True, # no autoclose for keyboard maestro
+        auto_close: bool = True, # no auto_close for keyboard maestro
         *args, **kwargs
     ):
         '''EITHER PROVIDE 'cookies_id' OR  'cookies_folder_path'.
@@ -118,7 +119,7 @@ class Firefox:
            appropriate amount of time for the page or a part of page to load; so there is a module named
            expected_conditions.
         '''
-        self.autoclose=autoclose  # for destructor
+        self.auto_close=auto_close  # for destructor
         # force headless on linux
         if platform.system()=='Linux':
             headless=True
@@ -253,7 +254,7 @@ class Firefox:
             from seleniumrequests import Firefox
             self.driver = Firefox(
                 firefox_profile=profile,
-                firefox_options=options,
+                options=options,
                 service_log_path=service_log_path,
                 log_path=log_path,
                 firefox_binary=ff_binary,
@@ -263,7 +264,7 @@ class Firefox:
             if not wire:
                 self.driver = webdriver.Firefox(
                     firefox_profile=profile,
-                    firefox_options=options,
+                    options=options,
                     service_log_path=service_log_path,
                     log_path=log_path,
                     firefox_binary=ff_binary,
@@ -275,7 +276,7 @@ class Firefox:
                 }
                 self.driver = webdriver.Firefox(
                     firefox_profile=profile,
-                    firefox_options=options,
+                    options=options,
                     seleniumwire_options=seleniumwire_options,
                     service_log_path=service_log_path,
                     log_path=log_path,
@@ -286,6 +287,9 @@ class Firefox:
         if full_screen:
             self.driver.fullscreen_window()
 
+        if max_window:
+            self.driver.maximize_window()
+
         if extensions_folder_path is not None:
             try:
                 change_timezone_id = None
@@ -351,6 +355,12 @@ class Firefox:
         if url is not None: self.driver.get(url)
 
     # -------------------------------------------------------- Public methods -------------------------------------------------------- #
+    def install(self,xpi,signed=True):
+        """
+        signed: the package has been signed or not
+        """
+        self.driver.install_addon(xpi, temporary=(not signed))
+
     def wait(self,seconds=None):
         if seconds is None: seconds=self.wait_ajax
         time.sleep(seconds)
@@ -724,40 +734,63 @@ class Firefox:
     def move_to_element(self, element: Optional[WebElement]) -> bool:
         if not element:
             print('move_to_element: None element passed')
-
             return False
-
         try:
             actions = ActionChains(self.driver)
             actions.move_to_element(element).perform()
-
             return True
         except Exception as e:
             print('move_to_element', e)
-
             return False
 
     def switch_to_default(self):
+        # default is the first frame (frame 0)?
         self.driver.switch_to.default_content()
         return True
 
     def switch_to_frame(self, iframe: Optional[WebElement]) -> bool:
         if not iframe:
             print('switch_to_frame: None frame passed')
-
             return False
-
         try:
             self.driver.switch_to.default_content()
             time.sleep(1)
             self.driver.switch_to.frame(iframe)
-
             return True
         except Exception as e:
             print('switch_to_frame', e)
-
             return False
 
+    def frames(self):
+        self.driver.switch_to.default_content()
+        time.sleep(1)
+        n = self.driver.execute_script('return window.length')
+        res = []
+        for i in range(0,n):
+            self.driver.switch_to.default_content()
+            time.sleep(1)
+            self.driver.switch_to.frame(i)
+            name = self.driver.execute_script('return window.name')
+            res.append({'i':i,'name':name})
+        self.driver.switch_to.default_content()
+        return res
+
+    def find_frame(self,by,key,*args,**kwargs):
+        # first frame is 0
+        self.driver.switch_to.default_content()
+        time.sleep(1)
+        n = self.driver.execute_script('return window.length')
+        for i in range(0,n):
+            self.driver.switch_to.default_content()
+            time.sleep(1)
+            self.driver.switch_to.frame(i)
+            if self.find(by,key,*args,**kwargs) is not None:
+                name = self.driver.execute_script('return window.name')
+                self.driver.switch_to.default_content()
+                return({'i':i,'name':name})
+        self.driver.switch_to.default_content()
+        return False
+
     # returns x, y, w, h, max_x, max_y
     def get_element_coordinates(self, element) -> Tuple[int, int, int, int, int, int]:
         location = element.location
@@ -868,9 +901,9 @@ class Firefox:
 
     def execute_script(self, script: str, element: Optional[WebElement] = None) -> bool:
         try:
-            self.driver.execute_script(script, element) if element else self.driver.execute_script(script)
+            return self.driver.execute_script(script, element) if element else self.driver.execute_script(script)
 
-            return True
+            # return True
         except Exception as e:
             print('{}: {} - {}'.format(inspect.stack()[2][3], script, e))
 
@@ -963,8 +996,8 @@ class Firefox:
     # --------------------------------------------------------- Destructor ----------------------------------------------------------- #
 
     def __del__(self):
-        # no autoclose for keyboard maestro
-        if self.autoclose: 
+        # no auto_close for keyboard maestro
+        if self.auto_close: 
             try:
                 # prevent double quit if already manually called quit
                 if os.path.exists(self.driver.profile.path):
@@ -1020,63 +1053,58 @@ class Firefox:
                 find_func((by, key))
             )
             # jerry: alias enhanced
-            es.send=es.send_keys
+            # es.send=es.send_keys
             # https://stackoverflow.com/a/54662690/2292993
             from functools import partial
-            def _senddelay(
+            def _send(
                 self,
                 keys: str,
-                min_delay: float = 0.025,
-                max_delay: float = 0.25
+                delay=[0.025,0.25],
+                n=None
             ) -> None:
+                # keys could only be single str with many chars
+                # delay in seconds after each key press, 0, [0.025,0.25] <-random between
                 import random
+                if type(delay) not in [tuple,list]: delay=[delay,delay]
                 for key in keys:
                     self.send_keys(key)
-                    time.sleep(random.uniform(min_delay,max_delay))
-            def _sendsubmit(self,*value):
+                    time.sleep(random.uniform(delay[0],delay[1]))
+            def _sendsubmit(self,*value,n=None):
                 # send keys and then submit
-                self.senddelay(*value)
+                self.send(*value)
                 time.sleep(0.5)
                 self.submit()
-            @property
-            def _location(self):
+            def _getarea(self):
                 # https://stackoverflow.com/a/59347207/2292993
                 # Assume there is equal amount of browser chrome on the left and right sides of the screen.
                 canvas_x_offset = self.parent.execute_script("return window.screenX + (window.outerWidth - window.innerWidth) / 2 - window.scrollX;")
                 # Assume all the browser chrome is on the top of the screen and none on the bottom.
                 canvas_y_offset = self.parent.execute_script("return window.screenY + (window.outerHeight - window.innerHeight) - window.scrollY;")
                 # Get the element center.
-                element_location = (self.rect["x"] + canvas_x_offset + self.rect["width"] / 2,
-                                    self.rect["y"] + canvas_y_offset + self.rect["height"] / 2)
-                return element_location
-            def _moveclick(self,n=2):
+                element_area = (self.rect["x"] + canvas_x_offset,
+                                self.rect["y"] + canvas_y_offset,
+                                self.rect["width"],
+                                self.rect["height"])
+                return element_area
+            def _moveclick(self,radius=5,n=2,wait=0.25,*args,**kwargs):
                 """
-                n = -1
-                    use selenium method
                 n could be 0; 2 in case the first click only gets the focus
-                    use python automation
                 """
-                if n==-1:
-                    # self is the element, self.parent is driver
-                    actions = ActionChains(self.parent)
-                    actions.move_to_element(self).click().perform()
-                else:
-                    # https://stackoverflow.com/a/59347207/2292993
-                    # Assume there is equal amount of browser chrome on the left and right sides of the screen.
-                    canvas_x_offset = self.parent.execute_script("return window.screenX + (window.outerWidth - window.innerWidth) / 2 - window.scrollX;")
-                    # Assume all the browser chrome is on the top of the screen and none on the bottom.
-                    canvas_y_offset = self.parent.execute_script("return window.screenY + (window.outerHeight - window.innerHeight) - window.scrollY;")
-                    # Get the element center.
-                    element_location = (self.rect["x"] + canvas_x_offset + self.rect["width"] / 2,
-                                        self.rect["y"] + canvas_y_offset + self.rect["height"] / 2)
-                    import ez
-                    ez.move(element_location)
-                    ez.click(n)
-            es.senddelay=partial(_senddelay, es)
+                # # self is the element, self.parent is driver
+                # actions = ActionChains(self.parent)
+                # actions.move_to_element(self).click().perform()
+                import ez
+                ez.moveclick(self.getarea(), radius, n, wait, *args, **kwargs)
+            def _movesend(self,keys,delay=[0.025,0.25],radius=5,n=2,wait=0.25):
+                self.moveclick(radius,n,wait)
+                import time
+                time.sleep(wait)
+                self.send(keys,delay)
+            es.send=partial(_send, es)
             es.sendsubmit=partial(_sendsubmit, es)
-            es.location=partial(_location, es)
-            es.loc=es.location
+            es.getarea=partial(_getarea, es)
             es.moveclick=partial(_moveclick, es)
+            es.movesend=partial(_movesend, es)
             return es
         except:
             return None
